/*
 * NetWinder LED Driver
 * Sean 2000 seanm@netwinder.org
 */

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/miscdevice.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <asm/uaccess.h>

#define LED_MINOR 151


/* Read the current state of the leds */
static int nwled_read(struct file *file, char *buf,
		      size_t len, loff_t * ppos)
{
	int leds = gpio_read();
	put_user_ret(leds, (int *) buf, -EFAULT);
	return sizeof(int);
}


/* Update the current state of the leds */
static int nwled_write(struct file *file, const char *events,
		       size_t len, loff_t * ppos)
{
	char evt;
	int i;

	for (i = 0; i < len; ++i, ++events) {
		get_user_ret(evt, events, -EFAULT);
		leds_event(evt);
	}

	return len;
}


static int nwled_open(struct inode *inode, struct file *filp)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
	return 0;
}


static int nwled_release(struct inode *inode, struct file *filp)
{
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
	return 0;
}


static struct file_operations nwled_fops = {
	NULL,			/* lseek */
	nwled_read,
	nwled_write,
	NULL,			/* readdir */
	NULL,			/* select */
	NULL,			/* ioctl */
	NULL,			/* mmap */
	nwled_open,
	NULL,			/* flush */
	nwled_release,
};

static struct miscdevice nwled_misc_device = {
	LED_MINOR,
	"nwled",
	&nwled_fops,
};


/*
 * This function is called from misc.c at bootup if the driver is
 * compiled into the kernel, or from init_module.
 */
int nwled_init(void)
{
	if (!machine_is_netwinder())
		return -ENODEV;

	printk(KERN_INFO "NetWinder LED Driver\n");

	if (misc_register(&nwled_misc_device)) {
		printk(KERN_WARNING
		       "nwled: Couldn't register device 10, %d.\n",
		       LED_MINOR);
		return -EBUSY;
	}

	return 0;
}


#ifdef MODULE
int init_module(void)
{
	return nwled_init();
}


void cleanup_module(void)
{
	misc_deregister(&nwled_misc_device);
}

#endif /* MODULE */
