/*
 *  fs/partitions/msdos.c
 *
 *  Code extracted from drivers/block/genhd.c
 *  Copyright (C) 1991-1998  Linus Torvalds
 *  Re-organised Feb 1998 Russell King
 */

#include <linux/config.h>
#include <linux/fs.h>
#include <linux/genhd.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/blk.h>
#include <linux/ctype.h>

#include <asm/system.h>

#include "check.h"

#ifdef CONFIG_ARCH_S390
#include <asm/ebcdic.h>
#include "../../drivers/s390/block/dasd_types.h"

dasd_information_t **dasd_information = NULL;

typedef enum {
  ibm_partition_none = 0,
  ibm_partition_lnx1 = 1,
  ibm_partition_vol1 = 3,
  ibm_partition_cms1 = 4
} ibm_partition_t;

static ibm_partition_t
get_partition_type ( char * type )
{
        static char lnx[5]="LNX1";
        static char vol[5]="VOL1";
        static char cms[5]="CMS1";
        if ( ! strncmp ( lnx, "LNX1",4 ) ) {
                ASCEBC(lnx,4);
                ASCEBC(vol,4);
                ASCEBC(cms,4);
        }
        if ( ! strncmp (type,lnx,4) ||
             ! strncmp (type,"LNX1",4) )
                return ibm_partition_lnx1;
        if ( ! strncmp (type,vol,4) )
                return ibm_partition_vol1;
        if ( ! strncmp (type,cms,4) )
                return ibm_partition_cms1;
        return ibm_partition_none;
}

int ibm_partition(struct gendisk *hd, kdev_t dev, int first_sector, int first_part_minor)
{
	struct buffer_head *bh;
	ibm_partition_t partition_type;
	char type[5] = {0,};
	char name[7] = {0,};
        int di = MINOR(dev) >> hd->minor_shift;
        if ( ! get_ptable_blocksize(dev) )
		return 0;
	if ( ! dasd_information )
		return 0;
	if ( ( bh = bread( dev, 
			 dasd_information[di]->sizes.label_block, 
			 dasd_information[di]->sizes.bp_sector ) ) != NULL ) {
		strncpy ( type,bh -> b_data, 4);
		strncpy ( name,bh -> b_data + 4, 6);
        } else {
		return 0;
	}
	if ( (*(char *)bh -> b_data) & 0x80 ) {
		EBCASC(name,6);
	}
	switch ( partition_type = get_partition_type(type) ) {
	case ibm_partition_lnx1:
		printk ( "(LNX1)/%6s:",name);
		add_partition( hd, MINOR(dev) + 1, 
				  (dasd_information[di]->sizes.label_block + 1) <<
				   dasd_information[di]->sizes.s2b_shift,
				  (dasd_information [di]->sizes.blocks -
				    dasd_information[di]->sizes.label_block - 1) <<
				  dasd_information[di]->sizes.s2b_shift,0 );
		break;
	case ibm_partition_vol1:
		printk ( "(VOL1)/%6s:",name);
		break;
	case ibm_partition_cms1:
		printk ( "(CMS1)/%6s:",name);
		if (* (((long *)bh->b_data) + 13) == 0) {
			/* disk holds a CMS filesystem */
			add_partition( hd, MINOR(dev) + 1, 
				       (dasd_information [di]->sizes.label_block + 1) <<
				       dasd_information [di]->sizes.s2b_shift,
				       (dasd_information [di]->sizes.blocks -
					dasd_information [di]->sizes.label_block) <<
				       dasd_information [di]->sizes.s2b_shift,0 );
			printk ("(CMS)");
		} else {
		  /* disk is reserved minidisk */
                        long *label=(long*)bh->b_data;
			int offset = label[13];
			int size = (label[7]-1-label[13])*(label[3]>>9);
			add_partition( hd, MINOR(dev) + 1, 
				       offset << dasd_information [di]->sizes.s2b_shift,
				       size<<dasd_information [di]->sizes.s2b_shift,0 );
			printk ("(MDSK)");
		}
		break;
	case ibm_partition_none:
		printk ( "(nonl)/      :");
		add_partition( hd, MINOR(dev) + 1, 
				  (dasd_information [di]->sizes.label_block + 1) <<
				  dasd_information [di]->sizes.s2b_shift,
				  (dasd_information [di]->sizes.blocks -
				   dasd_information [di]->sizes.label_block - 1) <<
				  dasd_information [di]->sizes.s2b_shift,0 );
		break;
	}
	printk ( "\n" );
	bforget(bh);
	return 1;
}
#endif


